package ru.dbotthepony.mc.otm.worldgen.placement

import com.github.benmanes.caffeine.cache.Caffeine
import com.github.benmanes.caffeine.cache.Scheduler
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import it.unimi.dsi.fastutil.HashCommon
import net.minecraft.Util
import net.minecraft.core.BlockPos
import net.minecraft.core.SectionPos
import net.minecraft.util.RandomSource
import net.minecraft.world.level.ChunkPos
import net.minecraft.world.level.levelgen.placement.PlacementContext
import net.minecraft.world.level.levelgen.placement.PlacementModifier
import ru.dbotthepony.mc.otm.core.util.CMWCRandom
import ru.dbotthepony.mc.otm.core.util.Xoshiro256SSRandom
import ru.dbotthepony.mc.otm.data.codec.inRange
import ru.dbotthepony.mc.otm.data.codec.minRange
import java.time.Duration
import java.util.stream.Stream
import kotlin.math.sqrt

/**
 * Enormous placement base, which allows it to span over several chunks without issues.
 *
 * MUST come as first placement modifier, other placement modifiers (such as rarity and
 * shuffle of center point within chunks) must be provided inside [Parameters.placementModifiers] list, in same order as if they were
 * *before* this placement
 */
abstract class AbstractEnormousPlacement(val parameters: Parameters) : PlacementModifier() {
	data class Parameters(
		/**
		 * How many chunks away to look for actual placements
		 *
		 * Too small value will cause placement cutoffs
		 */
		val chunkScanRange: Int,
		val seedMix: Long,

		/**
		 * Baseline placement modifiers, dictating how to appear in chunk
		 */
		val placementModifiers: List<PlacementModifier>,
	)

	private class GeneratedChunk(positions: Stream<BlockPos>) {
		// TODO: memory inefficient
		private val positions = ArrayList<BlockPos>()

		init {
			positions.forEach { this.positions.add(it) }
		}

		// TODO: this is primitive implementation, need to implement better one
		fun getPositions(pos: ChunkPos): Stream<BlockPos> {
			return positions.stream().filter {
				SectionPos.blockToSectionCoord(it.x) == pos.x &&
				SectionPos.blockToSectionCoord(it.z) == pos.z
			}
		}
	}

	protected abstract fun getPositions(center: BlockPos, random: RandomSource): Stream<BlockPos>

	private val chunkCache = Caffeine.newBuilder()
		.scheduler(Scheduler.systemScheduler())
		.executor(Util.backgroundExecutor())
		.maximumSize(16384L)
		.expireAfterWrite(Duration.ofMinutes(5))
		.softValues()
		.build<ChunkPos, GeneratedChunk>()

	private fun computeChunk(context: PlacementContext, pos: ChunkPos): GeneratedChunk {
		val random = Xoshiro256SSRandom(context.level.seed, parameters.seedMix, pos.x.toLong(), pos.z.toLong())
		var stream = Stream.of(BlockPos(pos.minBlockX, 0, pos.minBlockZ))
		parameters.placementModifiers.forEach { modifier -> stream = stream.flatMap { modifier.getPositions(context, random, it).sequential() } }
		return GeneratedChunk(stream.flatMap { getPositions(it, random) })
	}

	final override fun getPositions(context: PlacementContext, random: RandomSource, pos: BlockPos): Stream<BlockPos> {
		val cPos = ChunkPos(pos)
		val instances = ArrayList<GeneratedChunk>()

		for (x in -parameters.chunkScanRange .. parameters.chunkScanRange) {
			for (z in -parameters.chunkScanRange .. parameters.chunkScanRange) {
				// floor, so chunk scan range of 1 will give square instead of diamond
				val radius = sqrt(x.toDouble() * x + z.toDouble() * z).toInt()

				if (radius <= parameters.chunkScanRange) {
					val thisPos = ChunkPos(cPos.x + x, cPos.z + z)

					instances.add(chunkCache.get(thisPos) { computeChunk(context, thisPos) })
				}
			}
		}

		return instances.stream().flatMap { it.getPositions(cPos) }
	}

	companion object {
		val PARAMETERS_CODEC: MapCodec<Parameters> = RecordCodecBuilder.mapCodec {
			it.group(
				Codec.INT.minRange(0).fieldOf("chunk_scan_range").forGetter(Parameters::chunkScanRange),
				Codec.LONG.validate { if (it == 0L) DataResult.error { "Due to technical limitations, seed_mix can not be zero" } else DataResult.success(it) }.fieldOf("seed_mix").forGetter(Parameters::seedMix),
				CODEC.listOf().fieldOf("placement").forGetter(Parameters::placementModifiers),
			).apply(it, ::Parameters)
		}
	}
}
